//
// N A M E : d u m p _ p a c k e t . C
//
// D E S C : packet dump routine
//
// A U T H : Wataru Nishida, M.D., Ph.D.
//           wnishida@skyfree.org
//           http://www.skyfree.org/
//
// V E R S : It's still preliminary...
//
// D A T E : Mar 20th, 2001
//
// L  O  G : 2001/03/20  opened with sniff_***.C
//

#include <net/ethernet.h>	// ETHERTYPE_
#include <netinet/in.h>		// ntohs(), ntohl()

#include <stdio.h>		// sprintf()
#include <string.h>		// strcpy()

//
// class
//
//   M A C a d d r
//
//   --- MAC address information
//					Mar. 18, 2001

class MACaddr {
public:
  unsigned char addr[ 6 ];
public:
  const char* toString(void);
 };

const char* MACaddr::toString(void) {
  static char str[ 3*6 + 1 ];
  int i;
  char* ptr = str;

  for (i = 0; i < 5; i++) {
    sprintf(ptr, "%02X:", addr[ i ]);
    ptr += 3;
   }
  sprintf(ptr, "%02X", addr[ i ]);
  return(str);
 }

//
// class
//
//   I P v 4 a d d r
//
//   --- IPv4 address information
//					Mar. 18, 2001

class IPv4addr {
public:
  unsigned char addr[ 4 ];
public:
  unsigned long int toValue(void) \
    { return(ntohl(*((unsigned long int*) addr))); }
  const char* toString(int);
 };

const char* IPv4addr::toString(int pack) {
  static char str[ 4*4 + 1 ];
  int i;
  char  tmp[ 5 ];
  char* ptr = str;

  str[ 0 ] = 0;
  for (i = 0; i < 3; i++) {
    if (pack)
      sprintf(tmp, "%d.", this->addr[ i ]);
    else
      sprintf(tmp, "%03d.", this->addr[ i ]);
    strcat(ptr, tmp);
    ptr += strlen(tmp);
   }
  if (pack)
    sprintf(ptr, "%d", this->addr[ i ]);
  else
    sprintf(ptr, "%03d", this->addr[ i ]);
  return(str);
 }

//
// class
//
//   E h e a d e r
//
//   --- Ethernet frame header information
//					Mar. 18, 2001

#define DIXheader Eheader

class DIXheader {
public:
  MACaddr dst;
  MACaddr src;
  unsigned short int ftype;
public:
  unsigned int type(void) { return(ntohs(ftype)); }
 };

//
// class
//
//   A R P h e a d e r
//
//   --- ARP header information
//					Mar. 18, 2001

class ARPheader {
private:
  unsigned short int hard;	// Hardware address type 1:DIX, 6:IEEE
  unsigned short int prot;	// Protcol address type
  unsigned char haddlen;	// Hardware address length;
  unsigned char paddlen;	// Protocol address length;
  unsigned short int opcode;	// Operation code 1:ARP req, 2:ARP res,
                                //                3:RARP req, 4:RARP res
public:
  MACaddr  hsrc;		// Source hardware address
  IPv4addr psrc;		// Source protocol address
  MACaddr  hdst;		// Destination hardware address
  IPv4addr pdst;		// Destination protocol address

public:
  int htype(void) { return(ntohs(hard)); }
  int ptype(void) { return(ntohs(prot)); }
  int hlen(void) { return(haddlen); }
  int plen(void) { return(paddlen); }
  int ope(void) { return(ntohs(opcode)); }
 };

//
// class
//
//   I P h e a d e r
//
//   --- IP header information
//					Mar. 18, 2001

class IPheader {
private:
  unsigned char verlen;		// Version / Internet Header Length
public:
  unsigned char tos;		// Type of service
private:
  unsigned short int datalen;	// Datagram total length
  unsigned short int dataid;	// Datagram ID number
  unsigned short int flgoff;	// Flags / fragment offset
public:
  unsigned char ttl;		// Time to live
  unsigned char protocol;	// Protocol number
private:
  unsigned short int csum;	// Header checksum
public:
  IPv4addr src;			// Source address
  IPv4addr dst;			// Destination address

public:
  int version(void) { return((verlen & 0xF0) >> 4); }
  int headlen(void) { return((verlen & 0x0F) * 4); }
  int length(void) { return(ntohs(datalen)); }
  int id(void) { return(ntohs(dataid)); }
  int flags(void) { return((ntohs(flgoff) & 0xE000) >> 13); }
  int offset(void) { return(ntohs(flgoff) & 0x1FFF); }
  int chksum(void) { return(ntohs(csum) & 0xFFFF); }
 };

//
//   d u m p _ p a c k e t
//
//   --- packet dump routine
//					Mar. 18, 2001

void dump_packet(const char* buf, int len) {
  int type;
  Eheader* eh;
  ARPheader* ah;
  IPheader* ih;
  char src[32], dst[32];

  eh = (Eheader*) buf;
  type = eh->type();
  printf("=== Ethernet Header ===\n");
  strcpy(src, eh->src.toString());
  strcpy(dst, eh->dst.toString());
  printf("  Type %4X : MAC addr %s --> %s\n", type, src, dst);

  if (type == ETHERTYPE_ARP) {
    ah = (ARPheader*) (buf + sizeof(Eheader));
    printf("=== ARP Header ===\n");
    printf("  Operation code %d\n", ah->ope());
    printf("  Hardware address) type %6d : length %d\n", ah->htype(), ah->hlen());
    printf("  Protocol address) type 0x%04X : length %d\n", ah->ptype(), ah->plen());
    printf("  %s --> %s\n", ah->hsrc.toString(), ah->hdst.toString());
    printf("    %s --> %s\n", ah->psrc.toString(0), ah->pdst.toString(0));
   }
  else if (type == ETHERTYPE_IP) {
    ih = (IPheader*) (buf + sizeof(Eheader));
    printf("=== IP Header ===\n");
    printf("  Version %d : Hedaer length %d : TOS %d\n", ih->version(), ih->headlen(), ih->tos);
    printf("  Total Length %d : ID %d : offset %d\n", ih->length(), ih->id(), ih->offset());
    printf("  Protocol %d : TTL %d : Flags %d : Check sum %04X\n", ih->protocol, ih->ttl, ih->flags(), ih->chksum());
    strcpy(src, ih->src.toString(0));
    strcpy(dst, ih->dst.toString(0));
    printf("    %s --> %s\n", src, dst);
   }
 }
